#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/if_tun.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <unistd.h>
#include <assert.h>
#include <signal.h>
#include <time.h>

/* Controller for PlanetLab
   This class contains the main loop of the proxy, which handles
   all the network I/O. It creates, and receives packets from, the TUN/TAP
   device, it sends and receives tunneled packets to and from other
   proxies, and it sends and receives generic UDP packets. */

#include "bbemu.h"

#define ETH_HDR_SIZE 14
#define HDR_SIZE 4

LinuxController::LinuxController(in_addr_t localIP, int port, ParameterFile *params) : Controller()
{
  this->localIP = localIP;
  tunnelPort = port;
  tunnelSock = -1;
  numListenPorts = 0;
  numUDPRegistrations = 0;
  tunTapHandle = -1;
  nextTimer = -1;
  usingTAP = false;
  receiveBufferSize = -1;
  writeTimestamps = false;
  changeMTU = false;
  doPings = false;
  lastWakeup = getTimeMicros();
  for (int i=0; i<MAX_INSTANCES; i++)
    instanceTimer[i] = -1;
    
  if (params->selectSection("general")) {
    if (params->existsSetting("receiveBufferSize"))
      receiveBufferSize = params->getAsInt("receiveBufferSize");
    if (params->existsSetting("writeTimestamps"))
      writeTimestamps = params->getAsBoolean("writeTimestamps");
    if (params->existsSetting("doPings"))
      doPings = params->getAsBoolean("doPings");
    if (params->existsSetting("changeMTU"))
      changeMTU = params->getAsBoolean("changeMTU");
  }
}

static char *ipdisp(int ip, char *buf)
{
  sprintf(buf, "%d.%d.%d.%d", (ip>>24)&0xFF, (ip>>16)&0xFF, (ip>>8)&0xFF, ip&0xFF);
  return buf;
}

void quit(int signum)
{
  /* Before we terminate, we close the log file */

  if (globalLogFile) {
    FILE *theFile = globalLogFile;
    globalLogFile = NULL;
    fclose(theFile);
  }
  
  exit(1);
}

extern void recalculateChecksum(unsigned char *payload, int len);

void LinuxController::addTimestamp(unsigned char *data, int size, bool hasIPheader, long long timestamp)
{
  if (!writeTimestamps)
    return;
    
  unsigned char *payload = data;
  int payloadLen = size;
  
  if (hasIPheader && (size>=28)) {
    payload = &data[28];
    payloadLen -= 28;
  }
  
  if ((payloadLen<16) || (payload[2] != 0xAF) || (payload[3] != 0xFE)) 
    return;
  
  int maxTimestamps = (payloadLen-8)>>3;
  if (payload[1] >= maxTimestamps)
    return;
    
  *(long long*)&payload[8+8*payload[1]] = htonll(timestamp);
  payload[1] ++;
 
  if (hasIPheader) 
    recalculateChecksum(data, size);
}

int LinuxController::openTunTap(char *tunName, in_addr_t *tunIP, in_addr_t *tunMask, unsigned char *tunMAC) 
{
  /* Open the TUN device */

  int tunHandle;
  if ((tunHandle = open("/dev/net/tun", O_RDWR)) < 0) 
    panic("Cannot open /dev/net/tun (code %d). Is the tun module loaded and the device file available?", tunHandle);

  struct ifreq ifr;
  memset(&ifr, 0, sizeof(ifr));
  ifr.ifr_flags = IFF_TUN | IFF_NO_PI;

  if (ioctl(tunHandle, TUNSETIFF, (void*) &ifr) < 0)
    panic("Cannot set TUN device interface (%s)", strerror(errno));

  strcpy(tunName, ifr.ifr_name);

  /* Determine the TUN's IP and MAC addresses */

  int sd = socket(PF_INET, SOCK_STREAM, 0);
  if (sd < 0)
    panic("Cannot allocate socket (%s)", strerror(errno));

  /* Get the current flags */
  if (ioctl(sd, SIOCGIFFLAGS, &ifr) == -1) 
    panic("Failed to get TUN flags (%s)", strerror(errno));

  /* if_up */
  ifr.ifr_flags |= IFF_UP;
  if (ioctl(sd, SIOCSIFFLAGS, &ifr) == -1) 
    panic("Failed to set TUN flags (%s)", strerror(errno));

  /* Find the appropriate IP and netmask */
  
  bool haveIP = false;
  unsigned int ip = 0;
  
  for (int i=0; i<numInstances; i++) {
    if (instance[i].whichNode == localNode) {
      if (!haveIP) {
        ip = instance[i].internalAddr;
        haveIP = true;
      }
    }
  }
  
  if (!haveIP)
    panic("No instances on this node?!?");
      
  unsigned int netmask = 0xFFFFFFFE;
  for (int i=0; i<numInstances; i++) {
    while ((ntohl(instance[i].internalAddr) & netmask) != (ntohl(ip) & netmask)) {
      netmask <<= 1;
      if ((netmask & 0xFF000000) != 0xFF000000)
        panic("Instance IPs on this node are in more than one /8 subnet!");
    }
  }

  *tunIP = ip;
  *tunMask = htonl(netmask);

  struct sockaddr_in addr;

  memset(&ifr, 0, sizeof(ifr));
  strcpy(ifr.ifr_name, tunName);
  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = *tunIP;
  memmove(&ifr.ifr_addr, &addr, sizeof(addr));
  
  if (ioctl(sd, SIOCSIFADDR, &ifr) == -1) 
    panic("Setting TUN device address failed (%s)", strerror(errno));

  addr.sin_addr.s_addr = *tunMask;
  memmove(&ifr.ifr_addr, &addr, sizeof(addr));
  
  if (ioctl(sd, SIOCSIFNETMASK, &ifr) == -1) 
    panic("Setting TUN device netmask failed (%s)", strerror(errno));

  if (ioctl(sd, SIOCGIFHWADDR, &ifr) == -1) 
    panic("Failed to get TUN MAC addresss (%s)", strerror(errno));

  memcpy(tunMAC, &ifr.ifr_hwaddr.sa_data, 6);  
  close(sd);

  return tunHandle;
}

void LinuxController::acceptUDPmessages(Policy *policy, in_addr_t srcip, int srcport, int dstport)
{
  if (numUDPRegistrations >= MAX_REGISTRATIONS)
    panic("Too many registrations for UDP messages");
    
  char buf1[256];
  log("Accepting UDP messages from %s:%d on port %d", ipdisp(srcip, buf1), srcport, dstport);
    
  /* This may be called by a policy if it wants to receive UDP packets
     on a certain port. We open the port and do some bookkeeping to make 
     sure it is checked regularly by the main loop. */
    
  int portIndex = -1;
  for (int i=0; i<numListenPorts; i++) {
    if (listenPort[i].port == dstport)
      portIndex = i;
  }
  
  if (portIndex < 0) {
    if (numListenPorts >= MAX_LISTEN_PORTS)
      panic("Too many UDP listen ports");
      
    portIndex = numListenPorts ++;
    listenPort[portIndex].port = dstport;
    listenPort[portIndex].fd = socket(PF_INET, SOCK_DGRAM, 0);

    if (listenPort[portIndex].fd < 0)
      panic("Cannot open UDP port %d", dstport);
    
    log("Binding to port %d", dstport);
    
    sockaddr_in addr;
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(dstport);
    addr.sin_addr.s_addr = localIP;
    if (bind(listenPort[portIndex].fd, (struct sockaddr*)&addr, sizeof(addr)) < 0)
      panic("Cannot bind to UDP port %d", dstport);
  }
  
  udpRegistration[numUDPRegistrations].srcIP = srcip;
  udpRegistration[numUDPRegistrations].srcPort = srcport;
  udpRegistration[numUDPRegistrations].destPort = dstport;
  udpRegistration[numUDPRegistrations].handler = policy;
  numUDPRegistrations ++;
}

void LinuxController::acceptUDPmessagesNoMore(Policy *policy, in_addr_t srcip, int srcport, int dstport)
{
  char buf1[256];

  log("No longer accepting UDP messages from %s:%d on port %d", ipdisp(srcip, buf1), srcport, dstport);
  
  for (int i=0; i<numUDPRegistrations; i++) {
    if ((udpRegistration[i].srcIP == srcip) && (udpRegistration[i].srcPort == srcport) &&
        (udpRegistration[i].destPort == dstport) && (udpRegistration[i].handler == policy)) {
      log("Cancelling UDP registration for %s:%d", ipdisp(srcip, buf1), srcport);
      udpRegistration[i] = udpRegistration[--numUDPRegistrations];
      return;
    }
  }
  
  log("WARNING: Cannot cancel UDP registration for %s:%d", ipdisp(ntohl(srcip), buf1), srcport);
}

void LinuxController::tunnelPacket(int fromInstance, int toInstance, unsigned char *data, int size)
{
  assert((0<=toInstance) && (toInstance<numInstances));
  assert((0<=instance[toInstance].whichNode) && (instance[toInstance].whichNode<numNodes));

  /* Policies call this when they want to tunnel a data packet from one instance
     to another. If the instance is on another node, we send it to that
     node's tunnel port, otherwise we deliver it directly. */

  if (instance[toInstance].whichNode != localNode) {
    *(short*)&data[size] = htons(fromInstance);
    *(short*)&data[size+2] = htons(toInstance);

    if (fromInstance != 0xFFFF) {
      log("Tunneling packet %d->%d (%d bytes)", fromInstance, toInstance, size);
      event("SREM %d %d %d\n", fromInstance, toInstance, size);
      event2("SREM", data, size);

      addTimestamp(data, size, true, getTimeMicros());
    }
    
    struct sockaddr_in addr;  
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(node[instance[toInstance].whichNode].tunnelPort);
    addr.sin_addr.s_addr = node[instance[toInstance].whichNode].externalAddr;
        
    int ret = sendto(tunnelSock, data, size + HDR_SIZE, 0, (struct sockaddr*)&addr, sizeof(addr));
    if (ret <= 0)
      panic("Cannot send packet through tunnel (errno=%d)", errno);
  } else {
    log("Tunneling packet %d->%d (%d bytes) [LOCAL]", fromInstance, toInstance, size);
    forwardToDownstream(fromInstance, toInstance, data, size);
  }
}

void LinuxController::deliverPacket(int fromInstance, int toInstance, unsigned char *data, int size)
{
  assert((0<=toInstance) && (toInstance<numInstances));
  assert(instance[toInstance].whichNode == localNode);
  assert((0<=size) && (size<65536));

  addTimestamp(data, size, true, getTimeMicros());

  if (data[0] == 0xAA) {
    log("Internal ping packet going from [instance%d] to [instance%d]", fromInstance, toInstance);
    short echoSize = ntohs(*(short*)&data[4]);
    if (echoSize > 0) {
      log("Echoing back %d bytes", echoSize);
      unsigned char response[echoSize];
      for (int i=1; i<echoSize; i++)
        response[i] = (i<size) ? data[i] : 0;
      response[0] = 0xBB;
      instance[toInstance].policy->upstreamPacket(fromInstance, response, echoSize);
    }
    
    return;
  }

  /* Policies call this to deliver a packet to a local application instance. We do this
     simply by writing it to the TP after prefixing it with an Ethernet header. */

  if (usingTAP) {
    log("TAP delivery: %d->%d (%d bytes)", fromInstance, toInstance, size);
    event("STAP %d %d %d\n", fromInstance, toInstance, size);
    event2("STAP", data, size);

    unsigned char buf[size+14];
    memcpy(&buf[14], data, size);
    memcpy(&buf[0], tapMAC, 6);
    for (int i=0; i<6; i++)
      buf[6+i] = 0;
    buf[12] = 0x08;
    buf[13] = 0x00;

    int ret = write(tunTapHandle, buf, size+14);
    if (ret < 0)
      panic("Cannot send packet over TAP (errno=%d)", errno);
  } else {
    log("TUN delivery: %d->%d (%d bytes)", fromInstance, toInstance, size);
    
    int ret = write(tunTapHandle, data, size);
    if (ret < 0)
      panic("Cannot send packet over TUN (errno=%d)", errno);
  }
}

void LinuxController::forwardToDownstream(int fromInstance, int toInstance, unsigned char *data, int size)
{
  assert((0<=toInstance) && (toInstance<numInstances));
  
if (instance[toInstance].whichNode != localNode)
  panic("F%d T%d WN%d LN%d\n", fromInstance, toInstance, instance[toInstance].whichNode, localNode);
  assert(instance[toInstance].whichNode == localNode);
  
  instance[toInstance].policy->downstreamPacket(fromInstance, data, size);
}

void LinuxController::sendUDPMessage(int fromPort, in_addr_t toIP, int toPort, unsigned char *data, int size)
{
  int idx = -1;
  for (int i=0; i<numListenPorts; i++) {
    if (listenPort[i].port == fromPort)
      idx = i;
  }
  
  /* Policies can call this to send an UDP message to some IP/port */
  
  if (idx < 0)
    panic("Trying to send a UDP message from port %d, but no listeners are registered for that port", fromPort);
    
  struct sockaddr_in addr;  
  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(toPort);
  addr.sin_addr.s_addr = htonl(toIP);

  addTimestamp(data, size, false, getTimeMicros());

  int ret = sendto(listenPort[idx].fd, data, size, 0, (struct sockaddr*)&addr, sizeof(addr));
  if (ret <= 0)
    panic("Cannot send (errno=%d)", errno);
}

void LinuxController::setTimer(int instance, long long timeMicros)
{
  assert((0<=instance) && (instance<numInstances));

  instanceTimer[instance] = timeMicros;

  nextTimer = -1;
  for (int i=0; i<numInstances; i++) {
    if (instanceTimer[i] >= 0)
      if ((nextTimer < 0) || (instanceTimer[i] < nextTimer))
        nextTimer = instanceTimer[i];
  }
}

int LinuxController::maybeModifyPacket(unsigned char *buffer, int size)
{
  if (!changeMTU)
    return size;
  
  /* Is this an IPv4 packet? */
  
  if ((buffer[0]&0xF0) != 0x40)
    return size;
    
  int ihl = (buffer[0]&0x0F)*4;
  if ((ihl < 20) || (ihl > size))
    return size;
  
  /* Is this a TCP packet? */
    
  if (buffer[9] != 6)
    return size;
  
  unsigned char *tcpHdr = &buffer[ihl];
  int tcphl = 4*((tcpHdr[12]&0xF0) >> 4);
  if (size < (ihl+tcphl))
    return size;
    
  /* Is this a SYN packet? */
  
  if ((tcpHdr[13]&0x02) != 0x02)
    return size;
    
  int optionBytes = tcphl - 20;

  log("SYN found");

  /* If we already have a MSS option, find it */

  unsigned char *mssOption = NULL;
  unsigned char *thisOption = &tcpHdr[20];
  int optionBytesLeft = optionBytes;
  while (optionBytesLeft > 0) {
    int type = thisOption[0];
    if (type == 0)
      break;
      
    switch (type) {
      case 1 : /* NOP */
        thisOption += 1;
        optionBytesLeft --;
        break;
      case 2 : /* MSS */
        mssOption = thisOption;
        thisOption += 4;
        optionBytesLeft -= 4;
        break;
      case 3 : /* WSOPT */
        thisOption += 3;
        optionBytesLeft -= 3;
        break;
      case 4 : /* SACK permitted */
        thisOption += 2;
        optionBytesLeft -= 2;
        break;
      case 5 : /* SACK */
        if (thisOption[1] == 0)
          return size;
          
        optionBytesLeft -= thisOption[1];
        thisOption += thisOption[1];
        break;
      case 6 : /* ECHO */
        thisOption += 6;
        optionBytesLeft -= 6;
        break;
      case 7 : /* ECHO reply */
        thisOption += 6;
        optionBytesLeft -= 6;
        break;
      case 8 : /* TSOPT */
        thisOption += 10;
        optionBytesLeft -= 10;
        break;
      default :
        return size;
    }
  }
  
  /* No MSS option? Then we'll have to add one */
  
  if (!mssOption) {
    if (tcphl >= 60)
      return size;
      
    log("Adding MSS option");

    int tcpSize = size - ihl;
    for (int i=(tcpSize-1); i>=20; i--)
      tcpHdr[i+4] = tcpHdr[i];
      
    tcpHdr[12] += 0x10;
    *(unsigned short*)&buffer[2] = htons(ntohs(*(unsigned short*)&buffer[2]) + 4);
    size += 4;
    
    mssOption = &tcpHdr[20];
    mssOption[0] =  2;
    mssOption[1] = 4;
    *(unsigned short*)&mssOption[2] = htons(1500);
  }
  
  const static int maxMSS = 1400;
  int mss = ntohs(*(unsigned short*)&mssOption[2]);
  if (mss > maxMSS) {
    log("Changing MSS from %d to %d", mss, maxMSS);
    *(unsigned short*)&mssOption[2] = htons(maxMSS);
  }
    
  recalculateChecksum(buffer, size);
  return size;
}

int recvfrom_ts(int fd, unsigned char *buffer, int bufferSize, struct sockaddr *addr, socklen_t *addrLen, long long *recvTS)
{
  int retval = recvfrom(fd, buffer, bufferSize, 0, addr, addrLen);

  struct timeval tv;
    
  if (recvTS) {
    if (ioctl(fd, SIOCGSTAMP, &tv) >= 0)
      *recvTS = tv.tv_sec * 1000000ULL + tv.tv_usec;
    else
      *recvTS = getTimeMicros();
  }
    
  return retval;
}

void LinuxController::run()
{
  char buf1[200], buf2[200];

  if (geteuid() != 0)
    panic("Must be run as user 'root'");

  signal(SIGTERM, quit);
  signal(SIGINT, quit);
  signal(SIGSEGV, quit);

  /* Create TUN/TAP */

  char tapName[200];
  in_addr_t tapIP, tapMask;
  tunTapHandle = openTunTap(tapName, &tapIP, &tapMask, tapMAC);
  log("Opened %s device %s (IP %s, mask %s)", usingTAP ? "TAP" : "TUN", tapName, ipdisp(ntohl(tapIP), buf1), ipdisp(ntohl(tapMask), buf2));
  
  /* Open UDP sockets */

  tunnelSock = socket(PF_INET, SOCK_DGRAM, 0);
  if (tunnelSock < 0)
    panic("Cannot open tunneling socket");
    
  sockaddr_in addr;
  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(tunnelPort);
  addr.sin_addr.s_addr = localIP;
  if (bind(tunnelSock, (struct sockaddr*)&addr, sizeof(addr)) < 0)
    panic("Cannot bind to tunnel port %d : %s", tunnelPort, strerror(errno));

  if (receiveBufferSize >= 0) {
    if (setsockopt(tunnelSock, SOL_SOCKET, SO_RCVBUF, &receiveBufferSize, sizeof(receiveBufferSize)))
      panic("Cannot set receive buffer size on tunnel port to %d bytes", receiveBufferSize);

    int actualSize;
    socklen_t optlen2 = sizeof(actualSize);
    if (getsockopt(tunnelSock, SOL_SOCKET, SO_RCVBUF, &actualSize, &optlen2) < 0)
      panic("Cannot get actual receive buffer size on tunnel port");
      
    log("Set receive buffer size to %d bytes; actual size reported as %d bytes", receiveBufferSize, actualSize);
  }


  log("Bound to tunneling port %d", tunnelPort);

  /* Main loop */

  log("Entering main loop");
  
  long long lastFlush = getTimeMicros();
  long long nextPing = getTimeMicros();
  
  while (1) {
    fd_set fds;
    FD_ZERO(&fds);
    FD_SET(tunTapHandle, &fds);
    int maxfd = tunTapHandle;
    FD_SET(tunnelSock, &fds);
    if (tunnelSock > maxfd)
      maxfd = tunnelSock;
    for (int i=0; i<numListenPorts; i++) {
      FD_SET(listenPort[i].fd, &fds);
      if (listenPort[i].fd > maxfd)
        maxfd = listenPort[i].fd;
    }
    
    /* Determine the proper timeout */
    
    struct timeval to;
    struct timeval *pto = &to;
    long long now = getTimeMicros();
    if (nextTimer >= 0) {
      if (nextTimer < now)
        nextTimer = now;
        
      long long delay = nextTimer - now;
      to.tv_sec = (int)(delay/1000000LL);
      to.tv_usec = delay % 1000000LL;
    } else {
      pto = NULL;
    }
    
    /* Call select() */
    
    int ret = select(maxfd + 1, &fds, NULL, NULL, pto);
    if (ret < 0)
      panic("Select failed (errno=%d)", errno);
      
    /* See if any timers have expired */
    
    lastWakeup = getTimeMicros();
    now = lastWakeup;
    if (nextTimer <= now) {
      for (int i=0; i<numInstances; i++) {
        if ((instanceTimer[i]>=0) && (instanceTimer[i] < now)) {
          log("Timer #%d expired (delay %lldus)", i, now - instanceTimer[i]);
          setTimer(i, -1);
          instance[i].policy->timerExpired(now);
        }
      }
    }
      
    /* If something comes in over the TAP device, we give it to the policy, so it
       can emulate the upstream. */
      
    if (FD_ISSET(tunTapHandle, &fds)) {
      unsigned char packet[65536+ETH_HDR_SIZE];
      int size = read(tunTapHandle, packet, sizeof(packet));
      if (size <= 0)
        panic("Cannot read from TUN/TAP device");
        
      unsigned char *data = (unsigned char*) &packet[usingTAP ? ETH_HDR_SIZE : 0];
      char srcInfo[200], dstInfo[200];
      int fromInstance = -1;
      int toInstance = -1;

      addTimestamp(data, size, true, getTimeMicros());
      size = maybeModifyPacket(data, size);

      log("=== Received %s packet (%d bytes)", usingTAP ? "TAP" : "TUN", size);
      if (globalLogFile)
	dump(globalLogFile, packet, size);
      
      /* It is important to figure out which application instance this packet
         is a) coming from, and b) going to. We use both IP addresses and
         port numbers to distinguish between instances. Note than on 
         PlanetLab, binding to non-local IPs is forbidden, so we essentially
         only have the port numbers to go by. */
      
      if ((data[0]&0xF0) == 0x40) {
        in_addr_t fromIP = *(in_addr_t*)&data[12];
        in_addr_t toIP = *(in_addr_t*)&data[16];
        
        sprintf(srcInfo, "%s", ipdisp(ntohl(fromIP), buf1));
        sprintf(dstInfo, "%s", ipdisp(ntohl(toIP), buf1));
        if (((fromIP&tapMask)==(tapIP&tapMask)) && ((toIP&tapMask)==(tapIP&tapMask))) {
          int ihl = 4*(data[0]&0x0F);
          int proto = data[9];
          int fromPort = 0;
          int toPort = 0;
          if ((proto == 6) || (proto == 17)) {
            fromPort = ntohs(*(unsigned short*)&data[ihl]);
            toPort = ntohs(*(unsigned short*)&data[ihl+2]);
            sprintf(srcInfo, "%s:%d", ipdisp(ntohl(fromIP), buf1), fromPort);
            sprintf(dstInfo, "%s:%d", ipdisp(ntohl(toIP), buf1), toPort);
          }

          for (int i=0; i<numInstances; i++) {
            if ((instance[i].internalAddr == fromIP) && (fromPort >= instance[i].portRangeFrom) && (fromPort <= instance[i].portRangeTo))
              fromInstance = i;
            if ((instance[i].internalAddr == toIP) && (toPort >= instance[i].portRangeFrom) && (toPort <= instance[i].portRangeTo))
              toInstance = i;
          }
        }
      } else {
        strcpy(srcInfo, "[non-IP]");
        strcpy(dstInfo, "[non-IP]");
      }
      
      if ((fromInstance>=0) && (toInstance>=0)) {
        log("Intercepted packet [%s->%s] going from [instance%d] to [instance%d]", srcInfo, dstInfo, fromInstance, toInstance);
        assert((0<=fromInstance) && (fromInstance<numInstances) && (instance[fromInstance].whichNode == localNode));
        event("RTAP %d %d %d\n", fromInstance, toInstance, size);
        event2("RTAP", data, size - (usingTAP ? ETH_HDR_SIZE : 0));
        instance[fromInstance].policy->upstreamPacket(toInstance, data, size - (usingTAP ? ETH_HDR_SIZE : 0));
      } else {
        fprintf(stderr, "Unexpected packet: [%s->%s] (%d bytes)\n", srcInfo, dstInfo, size);
        log("Intercepted unexpected packet [%s->%s]", srcInfo, dstInfo);
        if (fromInstance < 0)
          log(" - Source [%s] does not match any instance; check IP and port range?", srcInfo);
        if (toInstance < 0)
          log(" - Destination [%s] does not match any instance; check IP and port range?", dstInfo);
        dump(stderr, &packet[0], size);
      }
    }
    
    /* If a packet comes in over the tunnel socket (from another node),
       we give it to the policy, so it can emulate the downstream. */
    
    if (FD_ISSET(tunnelSock, &fds)) {
      unsigned char packet[65536 + HDR_SIZE];
      struct sockaddr_in addr;
      socklen_t fromlen = sizeof(addr);
      long long recv_ts;
      int size = recvfrom_ts(tunnelSock, packet, sizeof(packet), (struct sockaddr*)&addr, &fromlen, &recv_ts);
      if (size <= 0)
        panic("Cannot read from tunnel");
        
      if ((size<4) || (*(int*)(&packet[size-4]) != 0xFFFFFFFF)) {
        addTimestamp(packet, size-4, true, recv_ts);
        addTimestamp(packet, size-4, true, getTimeMicros());

        log("=== Received TUNNEL packet (%d bytes)", size);
        if (globalLogFile)
	  dump(globalLogFile, packet, size);

        if (size >= HDR_SIZE) {
          short fromInstance = ntohs(*(short*)&packet[size - 4]);
          short toInstance = ntohs(*(short*)&packet[size - 2]);
          size -= 4;

          if ((0<=fromInstance) && (fromInstance<numInstances) && (0<=toInstance) && (toInstance<numInstances)) {
            if (instance[toInstance].whichNode == localNode) {
              event("RREM %d %d\n", fromInstance, toInstance, size);
              event2("RREM", packet, size);
              forwardToDownstream(fromInstance, toInstance, packet, size);
            } else {
              log("WARNING: Received a tunnel packet (%d->%d) that was meant for another node; discarding...", fromInstance, toInstance);
            }
          } else {
            log("Discarding invalid tunnel packet (%d->%d)", fromInstance, toInstance);
          }
        }
      } else {
        long long t1 = ntohll(*(long long*)(&packet[0]));
        long long t2 = ntohll(*(long long*)(&packet[8]));
        long long t3 = getTimeMicros(); 
        if (!t2) {
          *(long long*)(&packet[8]) = htonll(getTimeMicros());
          log("Received PING packet from %s:%d (t1=%lld, t2=%lld); turning around", ipdisp(ntohl(addr.sin_addr.s_addr), buf1), ntohs(addr.sin_port), t1, t3);
          int ret = sendto(tunnelSock, packet, size, 0, (struct sockaddr*)&addr, sizeof(addr));
          if (ret <= 0)
            panic("Cannot send pong packet (errno=%d)", errno);
        } else {
          log("Received PONG packet from %s:%d (t1=%lld, t2=%lld, t3=%lld)", ipdisp(ntohl(addr.sin_addr.s_addr), buf1), ntohs(addr.sin_port), t1, t2, t3);
          if (eventFile) {
            fprintf(eventFile, "%d.%06d XPING %s:%d %lld %lld %lld\n", 
              (int)(t3/1000000ULL), (int)(t3%1000000), ipdisp(ntohl(addr.sin_addr.s_addr), buf1), ntohs(addr.sin_port), t1, t2, t3
            ); 
          }
        }
      }
    }
    
    /* Check whether any packets came in on any of the UDP ports that 
       the policy asked us to listen to. If yes, we forward the
       packet to the policy for further processing. */
    
    for (int i=0; i<numListenPorts; i++) {
      if (FD_ISSET(listenPort[i].fd, &fds)) {
        unsigned char packet[65536];
        struct sockaddr_in addr;
        socklen_t fromlen = sizeof(addr);
        long long recv_ts;
        int size = recvfrom_ts(listenPort[i].fd, packet, sizeof(packet), (struct sockaddr*)&addr, &fromlen, &recv_ts);
        if (size <= 0)
          panic("Cannot read from UDP listen port");

        addTimestamp(packet, size, false, recv_ts);
        addTimestamp(packet, size, false, getTimeMicros());

        log("=== Received LISTEN packet (port %d, %d bytes)", listenPort[i].port, size);
	if (globalLogFile)
	  dump(globalLogFile, packet, size);

        in_addr_t srcIP = ntohl(addr.sin_addr.s_addr);
        int srcPort = ntohs(addr.sin_port);
        
        log("Incoming UDP message from %s:%d (%d bytes)", ipdisp(srcIP, buf1), srcPort, size);
        
        int idx = -1;
        for (int j=0; j<numUDPRegistrations; j++) {
          if ((udpRegistration[j].srcIP == srcIP) && (udpRegistration[j].srcPort == srcPort) &&
              (udpRegistration[j].destPort == listenPort[i].port)) {
            if (idx != -1)
              panic("Duplicate registration for UDP message from %X:%d", srcIP, srcPort);
              
            idx = j;
          }
        }

        if (idx < 0) {
          for (int j=0; j<numUDPRegistrations; j++) {
            if ((!udpRegistration[j].srcIP || (udpRegistration[j].srcIP == srcIP)) && 
                (!udpRegistration[j].srcPort || (udpRegistration[j].srcPort == srcPort)) &&
                (udpRegistration[j].destPort == listenPort[i].port)) {
              if (idx != -1)
                panic("Duplicate registration for UDP message from %X:%d", srcIP, srcPort);

              idx = j;
            }
          }
        }
        
        if (idx >= 0)
          udpRegistration[idx].handler->udpMessage(srcIP, srcPort, packet, size);
        else
          log("Nobody has registered for this packet; ignoring");
      }
    }
    
    /* If it has been too long since our last log flush, we do it now */
    
    if (lastWakeup>(lastFlush + 10000000)) {
      if (globalLogFile)
        fflush(globalLogFile);
      if (eventFile) 
        fflush(eventFile);
        
      lastFlush = lastWakeup;
    }
    
    if (lastWakeup >= nextPing) {
      if (doPings && (numNodes>1)) {
        int target;
        do {
          target = random() % numNodes;
        } while (target == localNode);

        log("Pinging node #%d", target);
        
        unsigned char buf[20];
        *(long long*)(&buf[0]) = htonll(getTimeMicros());
        *(long long*)(&buf[8]) = 0;
        *(int*)(&buf[16]) = 0xFFFFFFFF;

        struct sockaddr_in addr;  
        memset(&addr, 0, sizeof(addr));
        addr.sin_family = AF_INET;
        addr.sin_port = htons(node[target].tunnelPort);
        addr.sin_addr.s_addr = node[target].externalAddr;
        
        int ret = sendto(tunnelSock, buf, sizeof(buf), 0, (struct sockaddr*)&addr, sizeof(addr));
        if (ret <= 0)
          panic("Cannot send ping packet (errno=%d)", errno);
      }
      
      nextPing = lastWakeup + 2500000 + (random()%2500)*1000;
    }
  }
}
