#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <time.h>

#include "bbemu.h"
#include "parameters.h"

FILE *globalLogFile;
FILE *eventFile;
bool globalLogFlush;
int numNodes;
struct nodeInfo node[MAX_NODES];
int numInstances;
struct instanceInfo instance[MAX_INSTANCES];
int localNode;
bool planetlab;

int main(int argc, char *argv[])
{
  srandom(time(0));

  int argi = 1;
  while ((argi < argc) && (argv[argi][0] == '-')) {
    if (argv[argi][1] == 'c') {
      if (!strcmp(&argv[argi][2], "linux")) {
        planetlab = false;
      } else if (!strcmp(&argv[argi][2], "planetlab")) {
        planetlab = true;
      } else {
        panic("Unknown controller type: '%s'", &argv[argi][2]);
      }
    } else {
      panic("Unknown command line option: '%c'", argv[argi][1]);
    }
    
    argi ++;
  }
  
  if (argc != (argi + 2))
    panic("Syntax: %s [-c{linux,planetlab}] <parameterfile> <localNodeID>", argv[0]);
  
  /* Parse the local node ID. For example, if the local node
     is described under [node42] in the configuration file, 
     this ID should be 42. */
    
  localNode = atoi(argv[argi + 1]);
    
  /* If a log file is requested, open one */
    
  ParameterFile *params = new ParameterFile(argv[argi]);
  bool useNameServer = false;
  
  if (params->selectSection("general")) {
    if (params->existsSetting("logFile")) {
      char buf1[200];
      params->getAsString("logFile", buf1, sizeof(buf1));
      globalLogFile = fopen(buf1, "w+");
      if (!globalLogFile)
        panic("Cannot open global log file '%s'", buf1);
        
      if (params->existsSetting("logFlush"))
        globalLogFlush = params->getAsBoolean("logFlush");
      else
        globalLogFlush = false;
    }

    if (params->existsSetting("eventFile")) {
      char buf1[200];
      params->getAsString("eventFile", buf1, sizeof(buf1));
      eventFile = fopen(buf1, "w+");
      if (!eventFile)
        panic("Cannot open event file '%s'", buf1);
    }
    
    if (params->existsSetting("nameServerEnabled"))
      useNameServer = params->getAsBoolean("nameServerEnabled");
  }

  /* Read the node sections (e.g. [node0]). Each contains an
    'address=xxx' setting, where 'xxx' is the external, routable
    IP address of the node. */

  for (int i=0; i<MAX_NODES; i++) {
    char namebuf[200], valuebuf[200];
    sprintf(namebuf, "node%d", i);
    if (!params->selectSection(namebuf))
      break;
      
    numNodes = i+1;
    params->getAsString("address", valuebuf, sizeof(valuebuf));
    node[i].externalAddr = inet_addr(valuebuf);
    node[i].tunnelPort = params->getAsInt("tunnelPort");
  }
  
  if ((localNode<0) || (localNode>=numNodes))
    panic("Parameter file defines nodes 0..%d; localNodeID must be in that range", 0, numNodes-1);
  
  /* Instantiate the controller class, which does all the hard work */
  
  Controller *ctrl = planetlab ? new PlanetlabController(node[localNode].externalAddr, node[localNode].tunnelPort, params) : new LinuxController(node[localNode].externalAddr, node[localNode].tunnelPort, params);

  /* Read the instance descriptions (e.g. [instance37]). Each has
     an 'address=xxx' setting, where 'xxx' is the instance's local,
     non-routable address in the TUN/TAP network, a 'node=yyy'
     setting, where 'yyy' is the number of the node where the
     instance is to be located, and a 'type' setting, which
     is model' or 'passthrough'. */

  for (int i=0; i<MAX_INSTANCES; i++) {
    char namebuf[200], valuebuf[200];
    sprintf(namebuf, "instance%d", i);
    if (!params->selectSection(namebuf))
      break;
      
    numInstances = i+1;
    params->getAsString("address", valuebuf, sizeof(valuebuf));
    instance[i].internalAddr = inet_addr(valuebuf);
    instance[i].whichNode = params->getAsInt("node");
    if (params->existsSetting("ports")) {
      params->getAsString("ports", valuebuf, sizeof(valuebuf));
      char *pfrom = strtok(valuebuf, "-");
      char *pto = pfrom ? strtok(NULL, "-") : NULL;
      instance[i].portRangeFrom = atoi(pfrom);
      instance[i].portRangeTo = atoi(pto ? pto : pfrom);
    } else {
      instance[i].portRangeFrom = 0;
      instance[i].portRangeTo = 65535;
    }
    
    for (int j=0; j<i; j++) { 
      if ((instance[j].whichNode == instance[i].whichNode) && (instance[j].internalAddr == instance[i].internalAddr)) {
        if (((instance[i].portRangeFrom < instance[j].portRangeFrom) && (instance[i].portRangeTo >= instance[j].portRangeFrom)) ||
            ((instance[i].portRangeFrom >= instance[j].portRangeFrom) && (instance[i].portRangeFrom <= instance[j].portRangeTo)))
          panic("Port range overlap between instance%d and instance%d", i, j);
      }
    }
    
    if (instance[i].whichNode == localNode) {
      params->getAsString("type", valuebuf, sizeof(valuebuf));
      if (!strcasecmp(valuebuf, "model"))
        instance[i].policy = new ModelPolicy(ctrl, i, params);
      else if (!strcasecmp(valuebuf, "passthrough"))
        instance[i].policy = new PassThroughPolicy(ctrl, i, params);
      else
        panic("Unknown type '%s' specified for instance #%d", valuebuf, i);
    } else {
      instance[i].policy = NULL;
    }
  }

  /* The main loop is actually in the controller class */

  ctrl->run();
  delete ctrl;
  return 0;
}
