#include <stdlib.h>
#include <string.h>
#include <assert.h>

/* Model policy
   This policy uses a model of a broadband link (DSL or cable) to emulate the 
   link characteristics. The model is actually only for a unidirectional 
   broadband link, but we maintain two instances and parametrize them 
   differently for upstream and downstream. */

#include "bbemu.h"

ModelPolicy::ModelPolicy(Controller *controller, int instance, ParameterFile *params) : Policy(controller, instance)
{
  numSlotsUsed = 0;
  currentTimer = -1;
  upQueue = new BroadbandQueue();
  downQueue = new BroadbandQueue();

  char namebuf[200], valuebuf[200];
  sprintf(namebuf, "instance%d", instance);
  params->selectSection(namebuf);
  params->getAsString("model", valuebuf, sizeof(valuebuf));
  
  sprintf(namebuf, "model:%s", valuebuf);
  if (!params->selectSection(namebuf))
    panic("Cannot find model definition (section [%s]) in parameter file", namebuf);
  log("Created instance%d using %s", instance, namebuf);
    
  /* Link speed */
  
  downQueue->lineRateBps = params->getAsInt("downLinkSpeedBps");
  upQueue->lineRateBps = params->getAsInt("upLinkSpeedBps");
    
  /* Delay */
    
  downQueue->delayMicros = 0;
  upQueue->delayMicros = 0;
  if (params->existsSetting("rttMicros")) {
    upQueue->delayMicros = params->getAsInt("rttMicros") / 2;
    downQueue->delayMicros = upQueue->delayMicros;
  }
    
  /* Queue size */
  
  downQueue->capacityBytes = -1;
  if (params->existsSetting("downQueueSizeBytes"))
    downQueue->capacityBytes = params->getAsInt("downQueueSizeBytes");
  if (params->existsSetting("downQueueSizePackets"))
    downQueue->capacityPackets = params->getAsInt("downQueueSizePackets");

  upQueue->capacityBytes = -1;
  if (params->existsSetting("upQueueSizeBytes"))
    upQueue->capacityBytes = params->getAsInt("upQueueSizeBytes");
  if (params->existsSetting("upQueueSizePackets"))
    upQueue->capacityPackets = params->getAsInt("upQueueSizePackets");
}

int ModelPolicy::savePacket(int fromInstance, int toInstance, unsigned char *data, int size)
{
  while (numSlotsUsed && !packet[numSlotsUsed-1].inUse)
    numSlotsUsed --;
    
  int idx = -1;
  for (int i=0; (idx<0) && (i<numSlotsUsed); i++) {
    if (!packet[i].inUse)
      idx = i;
  }
  
  if (idx < 0) {
    if (numSlotsUsed == MAX_PACKETS)
      panic("Too many packets queued in ModelPolicy");
    
    idx = numSlotsUsed ++;
  }
  
  packet[idx].inUse = true;
  packet[idx].from = fromInstance;
  packet[idx].to = toInstance;
  packet[idx].size = size;
  memcpy(packet[idx].buf, data, size);
  log("Saved packet %d->%d (%d bytes) as P%d", fromInstance, toInstance, size, idx);
  return idx;
}

void ModelPolicy::releasePacket(int idx)
{
  assert((0<=idx) && (idx<MAX_PACKETS) && packet[idx].inUse);
  log("Released packet P%d", idx);
  packet[idx].inUse = false;
}

void ModelPolicy::updateTimer()
{
  long long t1 = upQueue->nextEvent();
  long long t2 = downQueue->nextEvent();
  long long newTimer = (t2>t1) ? t2 : t1;
  
  if (currentTimer != newTimer) {
    log("Changing timer %lld->%lld", currentTimer, newTimer);
    myController->setTimer(myInstance, newTimer);
    currentTimer = newTimer;
  }
}

void ModelPolicy::makeProgress(long long now)
{
  int id;
  while ((id = downQueue->getNextPacket(now)) >= 0) {
    myController->deliverPacket(packet[id].from, packet[id].to, packet[id].buf, packet[id].size);
    releasePacket(id);
  }

  while ((id = upQueue->getNextPacket(now)) >= 0) {
    myController->tunnelPacket(packet[id].from, packet[id].to, packet[id].buf, packet[id].size);
    releasePacket(id);
  }
}

void ModelPolicy::upstreamPacket(int toInstance, unsigned char *data, int size)
{
  long long now = myController->lastWakeupMicros();
  makeProgress(now);

  int id = savePacket(myInstance, toInstance, data, size);
  if (!upQueue->append(id, size, now)) {
    log("Dropped upstream packet");
    releasePacket(id);
    return;
  }    
  
  updateTimer();
}

void ModelPolicy::downstreamPacket(int fromInstance, unsigned char *data, int size)
{
  long long now = myController->lastWakeupMicros();
  makeProgress(now);
  
  int id = savePacket(fromInstance, myInstance, data, size);
  if (!downQueue->append(id, size, now)) {
    log("Dropped downstream packet");
    releasePacket(id);
    return;
  }    
  
  updateTimer();
}

void ModelPolicy::timerExpired(long long now)
{
  log("=== Timer expired at %lld", now);
  makeProgress(now);
  updateTimer();
}
